@@ -1,8 +1,8 @@
 //! A simple IRC crate written in rust
 //! ```no_run
-//! use circe::*;
+//! use circe::{commands::Command, Client, Config};
 //! fn main() -> Result<(), std::io::Error> {
-//!     let config = Config::from_toml("config.toml")?;
+//!     let config = Default::default();
 //!     let mut client = Client::new(config)?;
 //!     client.identify()?;
 //!
@@ -21,15 +21,22 @@
 //! }
 
 #![warn(missing_docs)]
+#![allow(clippy::too_many_lines)]
+#![cfg_attr(docsrs, feature(doc_cfg))]
+
+#[cfg(feature = "tls")]
 use native_tls::TlsConnector;
 
 use std::borrow::Cow;
-use std::fs::File;
 use std::io::{Error, Read, Write};
 use std::net::TcpStream;
-use std::path::Path;
 
+#[cfg(feature = "toml_support")]
 use serde_derive::Deserialize;
+#[cfg(feature = "toml_support")]
+use std::fs::File;
+#[cfg(feature = "toml_support")]
+use std::path::Path;
 
 /// IRC comamnds
 pub mod commands;
@@ -37,61 +44,77 @@ pub mod commands;
 /// An IRC client
 pub struct Client {
     config: Config,
-    stream: Option<TcpStream>,
-    sslstream: Option<native_tls::TlsStream<TcpStream>>,
+    #[cfg(not(feature = "tls"))]
+    stream: TcpStream,
+    #[cfg(feature = "tls")]
+    stream: native_tls::TlsStream<TcpStream>,
 }
 
 /// Config for the IRC client
-#[derive(Clone, Deserialize, Default)]
+#[derive(Clone, Default)]
+#[cfg_attr(feature = "toml_support", derive(Deserialize))]
 pub struct Config {
     channels: Vec<String>,
     host: String,
     mode: Option<String>,
     nickname: Option<String>,
     port: u16,
-    ssl: bool,
     username: String,
 }
 
+/// Custom Error for the `read` function
+#[derive(Debug)]
+pub struct NoNewLines;
+
+impl std::fmt::Display for NoNewLines {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(f, "Now new lines from the stream.")
+    }
+}
+
+impl std::error::Error for NoNewLines {}
+
 impl Client {
     /// Creates a new client with a given [`Config`].
     /// ```no_run
     /// # use circe::*;
-    /// # let config = Config::from_toml("config.toml")?;
+    /// # let config = Default::default();
     /// let mut client = Client::new(config)?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    ///
+    /// # Errors
     /// Returns error if the client could not connect to the host.
+    /// # Panics
+    /// Panics if the client can't connect to the given host.
     pub fn new(config: Config) -> Result<Self, Error> {
-        let stream = TcpStream::connect(format!("{}:{}", config.host, config.port))?;
+        let stream = TcpStream::connect(format!("{}:{}", config.host, config.port)).unwrap();
+        #[cfg(feature = "tls")]
         let sslstream: native_tls::TlsStream<TcpStream>;
 
-        if config.ssl {
+        #[cfg(feature = "tls")]
+        {
             let connector = TlsConnector::new().unwrap();
             sslstream = connector.connect(config.host.as_str(), stream).unwrap();
+        };
 
-            return Ok(Self {
-                config,
-                stream: None,
-                sslstream: Some(sslstream),
-            });
-        } else {
-            return Ok(Self {
-                config,
-                stream: Some(stream),
-                sslstream: None,
-            });
-        }
+        #[cfg(feature = "tls")]
+        return Ok(Self {
+            config,
+            stream: sslstream,
+        });
+
+        #[cfg(not(feature = "tls"))]
+        return Ok(Self { config, stream });
     }
 
     /// Identify user and joins the in the [`Config`] specified channels.
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.identify()?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
     /// Returns error if the client could not write to the stream.
     pub fn identify(&mut self) -> Result<(), Error> {
         self.write_command(commands::Command::CAP(commands::CapMode::LS))?;
@@ -105,7 +128,7 @@ impl Client {
         ))?;
 
         if let Some(nick) = self.config.nickname.clone() {
-            self.write_command(commands::Command::NICK(nick.to_string()))?;
+            self.write_command(commands::Command::NICK(nick))?;
         } else {
             self.write_command(commands::Command::NICK(self.config.username.clone()))?;
         }
@@ -117,6 +140,8 @@ impl Client {
                         self.write_command(commands::Command::PONG(code.to_string()))?;
                     }
                     commands::Command::OTHER(line) => {
+                        #[cfg(feature = "debug")]
+                        println!("{}", line);
                         if line.contains("001") {
                             break;
                         }
@@ -128,7 +153,7 @@ impl Client {
 
         let config = self.config.clone();
         self.write_command(commands::Command::MODE(config.username, config.mode))?;
-        for channel in config.channels.iter() {
+        for channel in &config.channels {
             self.write_command(commands::Command::JOIN(channel.to_string()))?;
         }
 
@@ -138,26 +163,24 @@ impl Client {
     fn read_string(&mut self) -> Option<String> {
         let mut buffer = [0u8; 512];
 
-        if self.config.ssl {
-            match self.sslstream.as_mut().unwrap().read(&mut buffer) {
-                Ok(_) => {}
-                Err(_) => return None,
-            };
-        } else {
-            match self.stream.as_mut().unwrap().read(&mut buffer) {
-                Ok(_) => {}
-                Err(_) => return None,
-            };
-        }
+        match self.stream.read(&mut buffer) {
+            Ok(_) => {}
+            Err(_) => return None,
+        };
 
-        Some(String::from_utf8_lossy(&buffer).into())
+        let res = String::from_utf8_lossy(&buffer);
+
+        // The trimming is required because if the message is less than 512 bytes it will be
+        // padded with a bunch of 0u8 because of the pre-allocated buffer
+        Some(res.trim().trim_matches(char::from(0)).trim().into())
     }
 
     /// Read data coming from the IRC as a [`commands::Command`].
     /// ```no_run
     /// # use circe::*;
+    /// # use circe::commands::Command;
     /// # fn main() -> Result<(), std::io::Error> {
-    /// # let config = Config::from_toml("config.toml")?;
+    /// # let config = Default::default();
     /// # let mut client = Client::new(config)?;
     /// if let Ok(ref command) = client.read() {
     ///     if let Command::OTHER(line) = command {
@@ -167,14 +190,18 @@ impl Client {
     /// # Ok::<(), std::io::Error>(())
     /// # }
     /// ```
+    /// # Errors
     /// Returns error if there are no new messages. This should not be taken as an actual error, because nothing went wrong.
-    pub fn read(&mut self) -> Result<commands::Command, ()> {
+    pub fn read(&mut self) -> Result<commands::Command, NoNewLines> {
         if let Some(string) = self.read_string() {
-            let command = commands::Command::from_str(&string);
+            #[cfg(feature = "debug")]
+            println!("{:#?}", string);
+
+            let command = commands::Command::command_from_str(&string);
 
             if let commands::Command::PONG(command) = command {
                 if let Err(_e) = self.write_command(commands::Command::PONG(command)) {
-                    return Err(());
+                    return Err(NoNewLines);
                 }
                 return Ok(commands::Command::PONG("".to_string()));
             }
@@ -182,7 +209,7 @@ impl Client {
             return Ok(command);
         }
 
-        Err(())
+        Err(NoNewLines)
     }
 
     fn write(&mut self, data: &str) -> Result<(), Error> {
@@ -191,15 +218,10 @@ impl Client {
             Cow::Owned(new) as Cow<str>
         };
 
-        if self.config.ssl {
-            self.sslstream
-                .as_mut()
-                .unwrap()
-                .write(formatted.as_bytes())
-                .unwrap();
-        } else {
-            self.stream.as_mut().unwrap().write(formatted.as_bytes())?;
-        }
+        #[cfg(feature = "debug")]
+        println!("{:?}", formatted);
+
+        self.stream.write_all(formatted.as_bytes())?;
 
         Ok(())
     }
@@ -208,13 +230,18 @@ impl Client {
     /// Not reccomended to use, use the helper functions instead.
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::PRIVMSG("".to_string() "#main".to_string(), "Hello".to_string()))?;
+    /// # use circe::commands::Command;
+    /// # let mut client = Client::new(Default::default())?;
+    /// client.write_command(Command::PRIVMSG("".to_string(), "#main".to_string(), "Hello".to_string()))?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
     /// Returns error if the client could not write to the stream.
     pub fn write_command(&mut self, command: commands::Command) -> Result<(), Error> {
-        use commands::Command::*;
+        use commands::Command::{
+            ADMIN, AWAY, CAP, INVITE, JOIN, LIST, MODE, NAMES, NICK, OPER, OTHER, PART, PASS, PING,
+            PONG, PRIVMSG, QUIT, TOPIC, USER,
+        };
         let computed = match command {
             ADMIN(target) => {
                 let formatted = format!("ADMIN {}", target);
@@ -225,7 +252,7 @@ impl Client {
                 Cow::Owned(formatted) as Cow<str>
             }
             CAP(mode) => {
-                use commands::CapMode::*;
+                use commands::CapMode::{END, LS};
                 Cow::Borrowed(match mode {
                     LS => "CAP LS 302",
                     END => "CAP END",
@@ -332,10 +359,12 @@ impl Client {
     /// Helper function for requesting information about the ADMIN of an IRC server
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.admin("192.168.178.100")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn admin(&mut self, target: &str) -> Result<(), Error> {
         self.write_command(commands::Command::ADMIN(target.to_string()))?;
         Ok(())
@@ -344,10 +373,12 @@ impl Client {
     /// Helper function for setting the users status to AWAY
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.away("AFK")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn away(&mut self, message: &str) -> Result<(), Error> {
         self.write_command(commands::Command::AWAY(message.to_string()))?;
         Ok(())
@@ -356,10 +387,12 @@ impl Client {
     /// Helper function for sending PRIVMSGs.
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.privmsg("#main", "Hello")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn privmsg(&mut self, channel: &str, message: &str) -> Result<(), Error> {
         self.write_command(commands::Command::PRIVMSG(
             String::from(""),
@@ -372,10 +405,12 @@ impl Client {
     /// Helper function to INVITE people to a channels
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.invite("liblirc", "#circe")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn invite(&mut self, username: &str, channel: &str) -> Result<(), Error> {
         self.write_command(commands::Command::INVITE(
             username.to_string(),
@@ -387,37 +422,29 @@ impl Client {
     /// Helper function for sending JOINs.
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.join("#main")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn join(&mut self, channel: &str) -> Result<(), Error> {
         self.write_command(commands::Command::JOIN(channel.to_string()))?;
         Ok(())
     }
 
-    /// Helper function for LISTing channels and modes
+    /// Helper function for ``LISTing`` channels and modes
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.list(None, None)?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn list(&mut self, channel: Option<&str>, server: Option<&str>) -> Result<(), Error> {
-        let channel_config = {
-            if let Some(channel) = channel {
-                Some(channel.to_string())
-            } else {
-                None
-            }
-        };
-        let server_config = {
-            if let Some(server) = server {
-                Some(server.to_string())
-            } else {
-                None
-            }
-        };
+        let channel_config = channel.map(std::string::ToString::to_string);
+        let server_config = server.map(std::string::ToString::to_string);
         self.write_command(commands::Command::LIST(channel_config, server_config))?;
         Ok(())
     }
@@ -425,10 +452,12 @@ impl Client {
     /// Helper function for getting all nicknames in a channel
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.names("#main,#circe", None)?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn names(&mut self, channel: &str, server: Option<&str>) -> Result<(), Error> {
         if let Some(server) = server {
             self.write_command(commands::Command::NAMES(
@@ -441,13 +470,32 @@ impl Client {
         Ok(())
     }
 
+    /// Helper function to try to register as a channel operator
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Default::default())?;
+    /// client.oper("username", "password")?;
+    /// # Ok::<(), std::io::Error>(())
+    /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
+    pub fn oper(&mut self, username: &str, password: &str) -> Result<(), Error> {
+        self.write_command(commands::Command::OPER(
+            username.to_string(),
+            password.to_string(),
+        ))?;
+        Ok(())
+    }
+
     /// Helper function for sending MODEs.
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.mode("test", Some("+B"))?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn mode(&mut self, target: &str, mode: Option<&str>) -> Result<(), Error> {
         if let Some(mode) = mode {
             self.write_command(commands::Command::MODE(
@@ -463,10 +511,12 @@ impl Client {
     /// Helper function for leaving channels.
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.part("#main")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn part(&mut self, target: &str) -> Result<(), Error> {
         self.write_command(commands::Command::PART(target.to_string()))?;
         Ok(())
@@ -475,10 +525,12 @@ impl Client {
     /// Helper function for setting or getting the topic of a channel
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.topic("#main", Some("main channel"))?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn topic(&mut self, channel: &str, topic: Option<&str>) -> Result<(), Error> {
         if let Some(topic) = topic {
             self.write_command(commands::Command::TOPIC(
@@ -494,10 +546,12 @@ impl Client {
     /// Helper function for leaving the IRC server and shutting down the TCP stream afterwards.
     /// ```no_run
     /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// # let mut client = Client::new(Default::default())?;
     /// client.quit(None)?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
     pub fn quit(&mut self, message: Option<&str>) -> Result<(), Error> {
         if let Some(message) = message {
             self.write_command(commands::Command::QUIT(message.to_string()))?;
@@ -507,14 +561,12 @@ impl Client {
                 env!("CARGO_PKG_VERSION")
             )))?;
         }
-        if self.config.ssl {
-            self.sslstream.as_mut().unwrap().shutdown().unwrap();
-        } else {
-            self.stream
-                .as_mut()
-                .unwrap()
-                .shutdown(std::net::Shutdown::Both)?;
-        }
+
+        #[cfg(not(feature = "tls"))]
+        self.stream.shutdown(std::net::Shutdown::Both)?;
+
+        #[cfg(feature = "tls")]
+        self.stream.shutdown()?;
 
         Ok(())
     }
@@ -526,7 +578,7 @@ impl Config {
     /// ```rust
     /// # use circe::*;
     /// let config = Config::new(
-    ///     vec!["#main", "#circe"],
+    ///     &["#main", "#circe"],
     ///     "192.168.178.100",
     ///     Some("+B"),
     ///     Some("circe"),
@@ -534,17 +586,20 @@ impl Config {
     ///     "circe",
     /// );
     /// ```
+    #[must_use]
     pub fn new(
-        channels: Vec<&'static str>,
+        channels: &[&'static str],
         host: &str,
         mode: Option<&'static str>,
         nickname: Option<&'static str>,
         port: u16,
-        ssl: bool,
         username: &str,
     ) -> Self {
         // Conversion from &'static str to String
-        let channels_config = channels.iter().map(|channel| channel.to_string()).collect();
+        let channels_config = channels
+            .iter()
+            .map(|channel| (*channel).to_string())
+            .collect();
 
         let mode_config: Option<String>;
         if let Some(mode) = mode {
@@ -566,17 +621,11 @@ impl Config {
             mode: mode_config,
             nickname: nickname_config,
             port,
-            ssl,
             username: username.into(),
         }
     }
 
     /// Create a config from a toml file
-    /// ```no_run
-    /// # use circe::*;
-    /// let config = Config::from_toml("config.toml")?;
-    /// # Ok::<(), std::io::Error>(())
-    /// ```
     /// ```toml
     /// channels = ["#main", "#main2"]
     /// host = "192.168.178.100"
@@ -585,7 +634,11 @@ impl Config {
     /// port = 6667
     /// username = "circe"
     /// ```
+    /// # Errors
     /// Returns an Error if the file cannot be opened or if the TOML is invalid
+    ///
+    #[cfg_attr(docsrs, doc(cfg(feature = "toml_support")))]
+    #[cfg(feature = "toml_support")]
     pub fn from_toml<P: AsRef<Path>>(path: P) -> Result<Self, Error> {
         let mut file = File::open(&path)?;
         let mut data = String::new();
